@@ -4,7 +4,7 @@ use Moo;
 use namespace::autoclean;
 use Types::Standard qw(InstanceOf Str Bool Maybe Int HashRef);
 
-use RDF::Trine qw[iri literal blank statement];
+use RDF::Trine qw[iri literal blank statement variable];
 use RDF::Trine::Serializer;
 use RDF::Trine::Namespace;
 use Log::Log4perl qw(:easy);
@@ -19,7 +19,10 @@ use RDF::RDFa::Generator 0.102;
 use HTML::HTML5::Writer qw(DOCTYPE_XHTML_RDFA);
 use Data::Dumper;
 use Digest::MD5 ('md5_base64');
+use Carp;
 use Try::Tiny;
+use List::Util qw(any);
+
 
 with 'MooX::Log::Any';
 
@@ -44,11 +47,11 @@ RDF::LinkedData - A simple Linked Data server implementation
 
 =head1 VERSION
 
-Version 0.69_05
+Version 0.69_06
 
 =cut
 
- our $VERSION = '0.69_05';
+ our $VERSION = '0.69_06';
 
 
 =head1 SYNOPSIS
@@ -94,13 +97,17 @@ L<Plack::Request> object (must be passed before you call C<content>)
 and an C<endpoint_config> hashref if you want to have a SPARQL
 Endpoint running using the recommended module L<RDF::Endpoint>.
 
-This module can also provide additional triples to turn the respons
+This module can also provide additional triples to turn the response
 into a hypermedia type. If you don't want this, set the C<hypermedia>
 argument to false. Currently this entails setting the SPARQL endpoint
 and vocabularies used using the L<VoID vocabulary|http://vocab.deri.ie/void>.
 The latter is very limited at present, all it'll do is use the namespaces
 if you have C<namespaces_as_vocabularies> enabled, which it is by default.
 
+Finally, it can provide experimental L<Triple Pattern
+Fragments|http://www.hydra-cg.com/spec/latest/triple-pattern-fragments/>
+support.
+
 =item C<< BUILD >>
 
 Called by Moo to initialize an object.
@@ -144,6 +151,9 @@ sub BUILD {
 		$self->void(RDF::Generator::Void->new(inmodel => $self->model, 
 														  dataset_uri => $dataset_uri,
 														  namespaces_as_vocabularies => $self->void_config->{namespaces_as_vocabularies}));
+		if ($self->has_fragments) {
+			$self->logger->debug('Triple Pattern Fragments config found with parameters: ' . Dumper($self->fragments_config) );
+		}
  	} else {
 		$self->logger->info('No VoID config found');
 	}
@@ -215,6 +225,8 @@ has endpoint_config => (is => 'rw',	isa=>Maybe[HashRef], predicate => 'has_endpo
 
 has void_config => (is => 'rw', isa=>Maybe[HashRef], predicate => 'has_void_config');
 
+has fragments_config => (is => 'rw', isa=>Maybe[HashRef], predicate => 'has_fragments');
+
 
 
 =item C<< request ( [ $request ] ) >>
@@ -299,6 +311,89 @@ sub response {
 	  }
 	}
 
+	if ($self->has_fragments && ($uri->path eq $self->fragments_config->{fragments_path})) {
+		croak 'A VoID description is needed when using Triple Pattern Fragments' unless ($self->has_void);
+		my %params = $uri->query_form;
+		my %statement = (subject => undef,
+							  predicate => undef,
+							  object => undef);
+		foreach my $term (keys(%statement)) {
+			my $value = $params{$term};
+			next unless $value;
+			return _client_error($response, "$term is invalid") if ref($value);
+			if ($value =~ m/^\?(\S+)$/) {
+				$statement{$term} = variable($1);
+			} elsif (($term eq 'object') && ($value =~ m/^\"(.+)\"((\@|\^\^)(\S+))?$/)) {
+				my $string = $1;
+				my $lang_or_datatype = $3;
+				my $rest = $4;
+				if (defined($lang_or_datatype) && ($lang_or_datatype eq '@')) {
+					$statement{$term} = literal($string, $rest);
+				} else {
+					$statement{$term} = literal($string, undef, $rest);
+				}
+			} else {
+				try {
+					$statement{$term} = iri($value);
+				} catch {
+					return _client_error($response, 'Was not able to parse subject as a IRI');
+				}
+			}
+		}
+
+		$self->logger->debug('Getting fragment with this selector ' . Dumper(\%statement));
+		return _client_error($response, 'Returning the whole database not allowed') unless any { defined } values(%statement);
+		my $output_model = $self->_common_fragments_control;
+
+		my $iterator = $self->model->get_statements($statement{subject}, $statement{predicate}, $statement{object});
+		$output_model->begin_bulk_ops;
+		my $counter = 0;
+		while (my $st = $iterator->next) {
+			$counter++;
+			# TODO: Paging goes here
+			$output_model->add_statement($st);
+		}
+		$self->add_namespace_mapping(void => 'http://rdfs.org/ns/void#');
+		$self->add_namespace_mapping(hydra => 'http://www.w3.org/ns/hydra/core#');
+		my $cl = literal($counter, undef, 'http://www.w3.org/2001/XMLSchema#integer');
+		my $void = RDF::Trine::Namespace->new('http://rdfs.org/ns/void#');
+		$output_model->add_statement(statement(iri($uri), 
+															$void->triples,
+															$cl));
+		$output_model->add_statement(statement(iri($uri), 
+															iri('http://www.w3.org/ns/hydra/core#totalItems'),
+															$cl));
+		$output_model->add_statement(statement(iri($uri),
+															iri('http://www.w3.org/1999/02/22-rdf-syntax-ns#type'),
+															$void->Dataset));
+		$output_model->add_statement(statement($self->void->dataset_uri,
+															$void->subset,
+															iri($uri)));
+		$output_model->end_bulk_ops;
+		my ($ct, $s);
+		try {
+			($ct, $s) = RDF::Trine::Serializer->negotiate('request_headers' => $headers_in,
+																		 base_uri => $self->base_uri,
+																		 namespaces => $self->_namespace_hashref);
+		} catch {
+			$response->status(406);
+			$response->headers->content_type('text/plain');
+			$response->body('HTTP 406: No serialization available any specified content type');
+			return $response;
+		};
+
+		$response->status(200);
+		$response->headers->header('Vary' => join(", ", qw(Accept)));
+		if (defined($self->last_etag)) {
+		  $response->headers->header('ETag' => '"' . $self->last_etag . '"');
+		}
+		my $body = $s->serialize_model_to_string($output_model);
+		$self->logger->trace("Fragment message body is $body");
+		$response->headers->content_type($ct);
+		$response->body(encode_utf8($body));
+		return $response;
+	}
+
 	if ($self->has_void) {
 		my $void_resp = $self->_void_content($uri, $endpoint_path);
 		return $void_resp if (defined($void_resp));
@@ -365,6 +460,15 @@ sub response {
 	return $response;
 }
 
+sub _client_error {
+	my ($response, $msg) = @_;
+	$response->status(400);
+	$response->headers->content_type('text/plain');
+	$response->body("HTTP 400: $msg");
+	return $response;
+}
+
+
 
 =item C<< helper_properties (  ) >>
 
@@ -527,7 +631,7 @@ has void => (is => 'rw', isa => InstanceOf['RDF::Generator::Void'], predicate =>
 sub _negotiate {
 	my ($self, $headers_in) = @_;
 	my ($ct, $s);
-	try {
+	eval {
 		($ct, $s) = RDF::Trine::Serializer->negotiate('request_headers' => $headers_in,
 																	 base_uri => $self->base_uri,
 																	 namespaces => $self->_namespace_hashref,
@@ -538,7 +642,7 @@ sub _negotiate {
 																	);
 		$self->logger->debug("Got $ct content type");
 		1;
-	} catch {
+	} or do {
 		my $response = Plack::Response->new;
 		$response->status(406);
 		$response->headers->content_type('text/plain');
@@ -613,6 +717,11 @@ sub _void_content {
 			$self->last_etag($self->current_etag);
 		}
 
+		if ($self->has_fragments) {
+			$self->add_namespace_mapping(hydra => 'http://www.w3.org/ns/hydra/core#');
+			$self->_common_fragments_control($self->_voidmodel);
+		}
+
 		# Now start serializing.
 		my ($ct, $s) = $self->_negotiate($self->request->headers);
 		return $ct if ($ct->isa('Plack::Response')); # A hack to allow for the failed conneg case
@@ -652,6 +761,53 @@ has _current_extvoid_mtime => (is => 'rw', isa => Int);
 
 has _last_extvoid_mtime => (is => 'rw', isa => Int);
 
+sub _common_fragments_control {
+	my $self = shift;
+	my $model = shift || RDF::Trine::Model->temporary_model;
+	my $void = RDF::Trine::Namespace->new('http://rdfs.org/ns/void#');
+	my $xsd  = RDF::Trine::Namespace->new('http://www.w3.org/2001/XMLSchema#');
+	my $hydra = RDF::Trine::Namespace->new('http://www.w3.org/ns/hydra/core#');
+	my $rdf = RDF::Trine::Namespace->new('http://www.w3.org/1999/02/22-rdf-syntax-ns#');
+	$model->begin_bulk_ops;
+	my $void_subject = $self->void->dataset_uri;
+	$model->add_statement(statement($void_subject,
+											  $rdf->type,
+											  $hydra->Collection));
+	$model->add_statement(statement($void_subject,
+											  $rdf->type,
+											  $void->Dataset));
+	$model->add_statement(statement($void_subject,
+											  $hydra->search,
+											  blank('template')));
+	$model->add_statement(statement($void_subject,
+								 $void->uriLookupEndpoint,
+								 literal($self->base_uri . $self->fragments_config->{fragments_path}
+											. '{?subject,predicate,object}')));
+	$model->add_statement(statement(blank('template'),
+								 $hydra->template,
+								 literal($self->base_uri . $self->fragments_config->{fragments_path}
+											. '{?subject,predicate,object}')));
+	$model->add_statement(statement(blank('template'),
+								 $hydra->property,
+								 $rdf->subject));
+	$model->add_statement(statement(blank('template'),
+								 $hydra->variable,
+								 literal('subject')));
+	$model->add_statement(statement(blank('template'),
+								 $hydra->property,
+								 $rdf->predicate));
+	$model->add_statement(statement(blank('template'),
+								 $hydra->variable,
+								 literal('predicate')));
+	$model->add_statement(statement(blank('template'),
+								 $hydra->property,
+								 $rdf->object));
+	$model->add_statement(statement(blank('template'),
+								 $hydra->variable,
+								 literal('object')));
+	$model->end_bulk_ops;
+	return $model;
+}
 
 =back
 
